"""LED board data"""

# pylint: disable=fixme
import logging
from neopolitan.board_functions.colors import ON, OFF

class Board:
    """Represents the colors for a board"""

    # pylint: disable=pointless-string-statement
    """
        board data:
            [ None, Red, Red, Red, None, Green, ... ]
    """

    def __init__(self, size):
        self.size = size
        self.data = []
        self.init_data()

    def init_data(self):
        """Initialize all 'lights' to 'off' (gray)"""
        self.data = [OFF for i in range(self.size)]

    def set_data(self, data, pad_to_end=True, cut_to_size=False):
        """Sets the board data and optionally makes it the right length"""

        # Pad the data so it fills the whole size
        if pad_to_end:
            if len(data) < self.size:
                data += [OFF for i in range(self.size - len(data))]
        # Cut the data so it doesn't "overflow" (even thought this does not cause errors)
        if cut_to_size:
            data = data[0:self.size]
        self.data = data

    def turn_on(self, idx, color=ON):
        """Set the color of a given idx (default=white)"""
        if idx >= len(self.data):
            # pylint: disable=consider-using-f-string
            logging.warning('index %s out of bounds: size = %s', idx, self.size)
            return
        self.data[idx] = color

    def turn_off(self, idx):
        """'Turn off' a given idx (set to None)"""
        # todo: what about with led board?
        self.data[idx] = OFF

    def set_color(self, idx, color):
        """Set the color of a given index"""
        self.turn_on(idx, color=color)

    def all_off(self):
        """Returns whether all the indices are off"""
        # would be much easier with the dictionary route
        for idx in self.data:
            if idx not in (None, OFF):
                return False
        return True

    def fill(self, color):
        """Sets all the colors in the board"""
        # todo: make better
        for i in range(self.size):
            self.turn_on(i, color)

    def scroll(self, wrap=True, height=8, pad=True):
        """'Scrolls' the data by one 'column' (height)"""
        if self.all_off():
            return # do nothing
        to_remove = self.data[0:height]
        self.data = self.data[height:]
        if wrap:
            self.data += to_remove
        if pad and len(self.data) < self.size:
            # pylint: disable=unused-variable
            for i in range(self.size - len(self.data)):
                self.data.append(OFF) # when to use None?



"""Tests whether the log can be written to"""

from neopolitan.log import init_logger, get_logger

def test_log():
    """Initializes the logger and writes to it"""

    # todo: doesn't do anything in pytest fixture
    init_logger()
    logger = get_logger()
    logger.debug('DEBUG')
    logger.info('INFO')
    logger.error('ERROR')



"""Tests various ways that board data is manipulated"""

from neopolitan.board_functions.board_data import default_board_data
from neopolitan.naples import process_board_data_events
from neopolitan.const import SCROLL_FAST

def test_board_data_event_processing():
    """Verify that processing events into board data object handles as expected"""

    # todo: more events

    board_data = default_board_data
    assert board_data is not None, 'Board data needs to be defined first'

    board_data = process_board_data_events(board_data, ['speed', 'fast'])
    assert board_data.scroll_speed == 'fast'\
        and board_data.scroll_wait == SCROLL_FAST,\
        'Scroll wait speed should be set to fast'

    board_data = process_board_data_events(board_data, ['speed', '0.7'])
    assert board_data.scroll_speed == 'user-defined'\
        and board_data.scroll_wait == 0.7,\
        'Scroll wait speed should be set to 0.7'

    board_data = process_board_data_events(board_data, ['speed', 'banana'])
    assert board_data.scroll_speed == 'user-defined'\
        and board_data.scroll_wait == 0.7,\
        'Scroll wait speed should not have changed'



"""Initialize the package logger"""

import logging
import datetime
import os
import sys

def init_logger(prep='logs/'):
    """Set up the log file"""
    # todo: prep arg validation
    # make directory if it does not exist
    # todo: is this ok?
    if not os.path.exists(prep):
        os.makedirs(prep)
    # todo: is all this formatting really necessary for a log file?
    log_time = str(datetime.datetime.now())
    filename = f'{os.getcwd()}/{prep}neopolitan {log_time}.txt'
    if sys.version_info[1] > 7:
        logging.basicConfig(filename=filename, encoding='utf=8', level=logging.DEBUG)
    else:
        logging.basicConfig(filename=filename, level=logging.DEBUG)

def get_logger():
    """Get the package logger"""
    return logging.getLogger('neopolitan')



"""
Main application code for displaying a board

Why is this file named 'naples.py'?
    To reduce the confusion of naming it 'neopolitan.py', I tried to find a synonym.
    Neopolitan/Neapolitan is a resident of Naples, Italy.
"""

import time

# pylint: disable=too-many-nested-blocks
# pylint: disable=too-many-branches
# pylint: disable=too-many-instance-attributes

from neopolitan.board_functions.board import Board
from neopolitan.writing.data_transformation import dispatch_str_or_lst, str_to_data
from neopolitan.board_functions.board_data import default_board_data
from neopolitan.log import get_logger
from neopolitan.const import WIDTH, HEIGHT


def process_board_data_events(board_data, event_list):
    """Manipulate board data according to events"""

    logger = get_logger()

    first = event_list[0]
    if first == 'speed':
        try:
            speed = event_list[1]
        # pylint: disable=broad-except
        except Exception as err:
            # todo: better explanation
            logger.warning('No second value provided, %s', err)
        if speed == 'slow':
            board_data.scroll_slow()
            logger.info('set speed: slow')
        elif speed == 'medium':
            board_data.scroll_medium()
            logger.info('set speed: medium')
        elif speed == 'fast':
            board_data.scroll_fast()
            logger.info('set speed: fast')
        else:
            try:
                speed = float(speed)
                board_data.set_scroll_wait(speed)
                logger.info('set speed: %s', speed)
            except ValueError:
                logger.warning('Cannot parse speed: %s', speed)
    elif first == 'wrap':
        try:
            wrap = event_list[1]
        # pylint: disable=broad-except
        except Exception as err:
            # todo: better explanation
            logger.warning('No second value provided, %s', err)
        if wrap in ('True', '1'):
            board_data.should_wrap = True
            logger.info('set wrap: True')
        elif wrap in ('False', '0'):
            board_data.should_wrap = False
            logger.info('set wrap: False')
        else:
            logger.warning('Cannot parse wrap: %s', wrap)

    return board_data

class Neopolitan:
    """Main application class for displaying a board"""
    def __init__(self, board_data=None, events=None):
        self.board_data = default_board_data if board_data is None else board_data
        self.width = WIDTH
        self.height = HEIGHT
        self.size = WIDTH*HEIGHT
        self.board = None
        self.display = None
        self.board_display = None

        self.events = events
        self.init_board()
        self.board.set_data(dispatch_str_or_lst(board_data.message))

    def __del__(self):
        del self.display

    def init_board(self):
        """Initialize board data"""
        # todo: make better
        if self.board_data.graphical:
            get_logger().info('Initializing graphical display')
            # pylint: disable=import-outside-toplevel
            from neopolitan.display.graphical_display import GraphicalDisplay
            self.board = Board(self.size)
            self.display = GraphicalDisplay(board=self.board)
        else:
            get_logger().info('Initializing hardware display')
            # pylint: disable=import-outside-toplevel
            from neopolitan.display.hardware_display import HardwareDisplay
            self.display = HardwareDisplay(WIDTH*HEIGHT)
            self.board_display = self.display.board_display
            self.board = self.board_display.board

    def loop(self):
        """Main display loop"""
        logger = get_logger()
        while not self.display.should_exit:
            # process events
            # todo: make better
            while self.events and not self.events.empty():
                event = self.events.get()
                logger.info('event: %s', event)
                event_list = event.split()
                first = event_list[0]
                if first == 'exit':
                    return
                if first == 'say':
                    logger.info('say: %s', event)
                    # todo: better handling: this is unintuitive
                    message = ' '
                    for word in event_list[1:]:
                        message += word + ' '
                    logger.info(message)
                    self.board.set_data(str_to_data(message))
                    logger.info('set message: %s', message)
                else: # try board data events
                    self.board_data = process_board_data_events(self.board_data, event_list)
                # todo: error handling
            self.display.loop()
            if self.board_data.scroll_speed:
                self.board.scroll(wrap=self.board_data.should_wrap)

            time.sleep(self.board_data.scroll_wait)



"""Functions for detecting the operating machine"""

import platform
from neopolitan.log import get_logger

def on_pi():
    """Is this code being run on a 32 bit Raspberry Pi?"""
    return (platform.machine()).strip() == 'armv7l'

def log_os():
    """Log whether this code is being run on a Pi"""
    get_logger().info('On Pi: %s', (platform.machine()).strip() == 'armv7l')
    get_logger().info(' Machine: %s', platform.machine())



"""Demo some functionalities of the board"""

from neopolitan.log import init_logger
from neopolitan.naples import Neopolitan
from neopolitan.board_functions.board_data import default_board_data
# pylint: disable=wildcard-import
# pylint: disable=unused-wildcard-import
from neopolitan.board_functions.colors import *
# pylint: disable=anomalous-backslash-in-string

LOWER = 'abcdefghijklmnopqrstuvwxyz'
UPPER = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
NUMBERS = '0123456789'
SYMBOLS = '$ % ↑ ↓ ( ) - . , : = ~ ! @ & * ? < > ; | { } " \''

def display(msg):
    """Display a message on the board"""

    init_logger()

    board_data = default_board_data
    board_data.message = msg

    # board_data.scroll_fast()

    neop = Neopolitan(board_data=board_data)
    neop.loop()
    del neop

def display_all():
    """Display all defined characters"""
    display(f'    {LOWER} {UPPER} {NUMBERS} {SYMBOLS}')

def display_all_lowercase_letters():
    """Display all lowercase letters"""
    display(LOWER)

def display_all_uppercase_letters():
    """Display all uppercase letters"""
    display(UPPER)

def display_all_numbers():
    """Display all numbers"""
    display(NUMBERS)

def display_all_symbols():
    """Display all symbols"""
    display(SYMBOLS)

def color_demo():
    """Displays a rainbow message"""
    msg = [
        ('-', WHITE),
        ('-', RED),
        ('-', ORANGE),
        ('-', YELLOW),
        ('-', GREEN),
        ('-', BLUE),
        ('-', PURPLE),
    ]
    display(msg)



"""Tests whether the 'flip' functionality is implemented correctly"""

# pylint: disable=import-error
from neopolitan.display.hardware_board_display import HardwareBoardDisplay

# pylint: disable=pointless-string-statement
"""
0 0 0 -> 0 1 0
1 1 1    1 0 1
"""

def test_flip():
    """Test that the flip functions as expected"""
    # todo: more data tests

    initial_data = [0,1,0,1,0,1]
    expected_data = [0,1,1,0,0,1]
    height = 2

    assert HardwareBoardDisplay.flip(initial_data, height=height) == expected_data, \
        'Flipped data does not match expected'



"""Displays a graphical representation of the LED board"""

import pygame
# pylint: disable=fixme
# pylint: disable=import-error
# pylint: disable=no-name-in-module
# todo: why import error??

from neopolitan.display.abstract_display import Display
from neopolitan.display.graphical_board_display import GraphicalBoardDisplay
from neopolitan.const import WIDTH, HEIGHT

# todo: cleanup pygame

class GraphicalDisplay(Display):
    """A graphical LED display"""
    # pylint: disable=too-many-instance-attributes

    # pylint: disable=too-many-arguments
    def __init__(
            self, board,
            rows=HEIGHT, cols=WIDTH,
            width=1600, height=400,
            background=(0, 20, 30), size=20, space=50):
        super().__init__()

        self.background = background
        self.width = width
        self.height = height
        self.board = board
        self.size = size
        self.space = space

        self.board_display = GraphicalBoardDisplay(board, cols, rows)
        self.init_pygame()

    def __del__(self):
        """Clean up pygame"""
        # pylint: disable=no-member
        pygame.quit()

    def init_pygame(self):
        """Initialize pygame"""
        # ToDo
        # pylint: disable=no-member
        pygame.init()
        self.screen = pygame.display.set_mode((self.width, self.height))
        # pygame.display.set_caption('stock board simulator')
        # icon = pygame.image.load('./assets/images/wsb.png')
        # pygame.display.set_icon(icon)

    def draw(self):
        """Main draw loop"""
        self.screen.fill(self.background)
        self.board_display.draw_board(self.screen, self.space, self.size)

    def loop(self):
        """Loop until 'esc' or quit"""
        for event in pygame.event.get():
            # ToDo
            # pylint: disable=no-member
            if event.type == pygame.QUIT:
                # pylint: disable=attribute-defined-outside-init
                self.should_exit = True
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_SPACE:
                    self.board.turn_on(0)
            if event.type == pygame.KEYUP:
                if event.key == pygame.K_SPACE:
                    self.board.turn_off(0)
        self.draw()
        pygame.display.update()
        return self.should_exit
    


"""Test groups"""

from neopolitan.writing.letters_8 import  \
    DEFINED_UPPERCASE, DEFINED_LOWERCASE, DEFINED_SYMBOLS, DEFINED_NUMBERS
from neopolitan.writing.groups_8 import  \
    uppercase, lowercase, symbols, numbers


def test_all_letters_used():
    """Tests that all defined symbols are put into a group"""
    len_groups = len(uppercase) + len(lowercase) + len(symbols) + len(numbers)
    len_letters = DEFINED_UPPERCASE + DEFINED_LOWERCASE + DEFINED_SYMBOLS + DEFINED_NUMBERS
    assert len_groups == len_letters, 'The number of symbols defined should be the same either way'
    


"""Demo test"""
from neopolitan import main

def demo_test():
    """Test demo"""
    try:
        main.main()
    # pylint: disable=broad-except
    except Exception as err:
        assert False, str(err)



"""Holds data for displaying a board"""
from dataclasses import dataclass
from neopolitan.const import SCROLL_SLOW, SCROLL_MED, SCROLL_FAST
from neopolitan.os_detection import on_pi
# pylint: disable=wildcard-import
# pylint: disable=unused-wildcard-import
from neopolitan.board_functions.colors import *

@dataclass
class BoardData: # todo: DisplayData?
    """Holds data for displaying a board"""
    message: str
    graphical: bool
    scroll_speed: str
    scroll_wait: float
    should_wrap: bool

    def scroll_fast(self):
        """Set scroll speed to fast"""
        self.scroll_speed = 'fast'
        self.scroll_wait = SCROLL_FAST

    def scroll_medium(self):
        """Set scroll speed to medium"""
        self.scroll_speed = 'medium'
        self.scroll_wait = SCROLL_MED

    def scroll_slow(self):
        """Set scroll speed to slow"""
        self.scroll_speed = 'slow'
        self.scroll_wait = SCROLL_SLOW

    def set_scroll_wait(self, timer):
        """Set scroll speed to a user-defined value"""
        # todo: handle ints too
        assert isinstance(timer, float), f'A float must be provided: {timer}'
        self.scroll_wait = timer
        self.scroll_speed = 'user-defined'

default_message = [
    ('              hello, and welcome to ', WHITE),
    ('n', RED),
    ('e', ORANGE),
    ('o', YELLOW),
    ('p', GREEN),
    ('o', BLUE),
    ('l', PURPLE),
    ('i', RED),
    ('t', ORANGE),
    ('a', YELLOW),
    ('n', GREEN),
    ('!', WHITE)
]
default_board_data = BoardData(default_message, not on_pi(), 'medium', SCROLL_MED, True)



"""Interacts with the LED board"""

import logging
from neopolitan.display.abstract_board_display import BoardDisplay
from neopolitan.const import WIDTH, HEIGHT

class HardwareBoardDisplay(BoardDisplay):
    """Draws board data"""

    def __init__(self, board, pixels, size=WIDTH*HEIGHT):
        # pylint: disable=line-too-long
        # pylint: disable=consider-using-f-string
        super().__init__(board)
        assert pixels, 'Neopixel library not initialized'
        self.pixels = pixels
        self.size = size # todo: organization
        assert board.size == self.size, \
            'board size ({0} does not meet given size ({1})'.format(board.size, size)

    def draw_board(self):
        """Sets all the LEDs in accordance with the current data"""
        assert self.board, 'No board assigned'

        flipped_data = HardwareBoardDisplay.flip(self.board.data, HEIGHT)

        for i in range(self.size):
            if i >= len(self.board.data):
                logging.warning('index %s outside of data array bounds', i)
                return
            self.pixels[i] = flipped_data[i]

    # pylint: disable=no-self-argument
    def flip(data, height=HEIGHT, start_at_first=False):
        """
        Handles flipping alternate 'rows' so that data appears as expected;
        returns the flipped data
        """

        # Every other 'column', starting with the second,
        # should be flipped such that it appears 'upside down',
        # so that when the board displays it actually appears right side up.
        assert len(data) % height == 0, \
            'Data length does not fill its last column' # todo: doesn't need to tho

        flipped_data = []
        idx = 0
        # pylint: disable=unsubscriptable-object
        while idx < len(data):
            col = data[idx:idx+height]
            if start_at_first:
                col = reversed(col)
            start_at_first = not start_at_first

            flipped_data += col
            idx += height

        return flipped_data

    # TODO: SET DATA TO MATCH!!
    def fill(self, color):
        """Fill all the pixels a given color"""
        # self.pixels.fill(color)
        self.board.fill(color)

    def fill_red(self):
        """Fill all the pixels red"""
        self.fill((255,0,0))
    def fill_green(self):
        """Fill all the pixels green"""
        self.fill((0,255,0))
    def fill_blue(self):
        """Fill all the pixels blue"""
        self.fill((0,0,255))
    def fill_white(self):
        """Fill all the pixels white"""
        self.fill((255,255,255))
    def fill_blank(self):
        """Turn all the pixels off"""
        # todo: unecessary
        self.fill((0,0,0))



"""Takes letters as defined and transforms them into usable data"""

# pylint: disable=fixme
# Todo: location?
# todo
from math import floor
import logging
from neopolitan.writing.groups_8 import uppercase, lowercase, symbols, numbers
from neopolitan.board_functions.colors import ON, OFF

def character_to_symbol(char):
    """Gets the symbol for the character"""
    ascii_val = ord(char)
    # todo: make everything like for symbols?
    if char == ' ':
        # todo
        return 'space'
    if 97 <= ascii_val < 123:
        return lowercase[ascii_val-97]
    if 65 <= ascii_val < 91:
        return uppercase[ascii_val-65]
    if 48 <= ascii_val < 58:
        return numbers[ascii_val-48]
    if char in symbols:
        return symbols[char]
    logging.warning('Cannot find char: %s', char)
    return []

def frame_length(sym):
    """Returns the highest multiple of 8 above the largest index in the symbol array.
    This makes it so the frame has the correct length, since it should 'fill' all columns it uses"""
    if sym is None or len(sym) == 0:
        return 0
    if sym == 'space':
        return 8*2
    sym_len = len(sym)
    last_idx = sym_len-1
    last_col = sym[last_idx]
    len_last_col = len(last_col)
    last_col_idx = len_last_col-1
    last_val = last_col[last_col_idx]
    # round up to next multiple of 8
    ret = -1
    if last_val % 8 == 0:
        ret = last_val
    else:
        ret = 8 * (floor(last_val / 8.0)+1)
    return ret

# Todo: dictionary: idx=color instead?
def symbol_to_array(sym, color=ON, off=None):
    """Takes a defined symbol and returns an array where the symbol
        is defined by indices of 'color' and 'off' values are None"""
    frame = [off for i in range(frame_length(sym))]
    if sym == 'space':
        # todo
        return frame
    for col in sym:
        for val in col:
            frame[val] = color
    return frame

def str_to_data(msg, color=ON, add_space=4):
    """Converts a string into usable display data"""

    data = []
    line = [OFF for i in range(8)] # todo: height?
    for char in msg:
        arr = symbol_to_array(character_to_symbol(char), color=color, off=OFF)
        data = data + arr + line
    data += line*add_space
    return data

def color_list_to_data(arr):
    """Converts a list of (message, color) objects into a colored board message"""
    data = []
    for msg, col in arr:
        data += str_to_data(msg, color=col, add_space=0)
    return data

def dispatch_str_or_lst(msg):
    """Dispatches given the input to either the basic-string or colored-string data makers"""
    # ToDo: this is not the most elegant system ever
    if isinstance(msg, str):
        return str_to_data(msg)
    if isinstance(msg, list):
        if len(msg) > 0:
            if isinstance(msg[0], tuple):
                return color_list_to_data(msg)
    return []



"""All groups of symbols for height of 8"""

# pylint: disable=fixme
# ToDo: fix this
# pylint: disable=wildcard-import
# pylint: disable=unused-wildcard-import
from neopolitan.writing.letters_8 import *

uppercase = [
    A,
    B,
    C,
    D,
    E,
    F,
    G,
    H,
    I,
    J,
    K,
    L,
    M,
    N,
    O,
    P,
    Q,
    R,
    S,
    T,
    U,
    V,
    W,
    X,
    Y,
    Z,
]

lowercase = [
    a,
    b,
    c,
    d,
    e,
    f,
    g,
    h,
    i,
    j,
    k,
    l,
    m,
    n,
    o,
    p,
    q,
    r,
    s,
    t,
    u,
    v,
    w,
    x,
    y,
    z,
]

symbols = {
    "$": DOLLAR,
    "%": PERCENT,
    "↑": UP,
    "↓": DOWN,
    "(": OPEN,
    ")": CLOSE,
    "-": MINUS,
    ".": PERIOD,
    ":": COLON,
    "=": EQUALS,
    "~": TILDE,
    "!": EXCLAMATION,
    "@": AT,
    "&": AMPERSAND,
    "*": ASTERISK,
    "?": QUESTION,
    "<": LESSTHAN,
    ">": GREATERTHAN,
    ";": SEMICOLON,
    "|": PIPE,
    "{": OPENCURLY,
    "}": CLOSECURLY,
    '"': DOUBLEQUOTE,
    "'": SINGLEQUOTE,
    ",": COMMA,
}

numbers = [ZERO, ONE, TWO, THREE, FOUR, FIVE, SIX, SEVEN, EIGHT, NINE]

def space(height=8):
    """Returns a space (empty column) of specified height"""
    return [[height-1]]



"""Main application function"""

# pylint: disable=too-many-nested-blocks
# pylint: disable=too-many-branches

import getopt
import sys

from neopolitan.os_detection import log_os
from neopolitan.log import init_logger, get_logger
from neopolitan.naples import Neopolitan
from neopolitan.os_detection import on_pi
from neopolitan.board_functions.board_data import default_board_data
from neopolitan.const import SCROLL_SLOW, SCROLL_MED, SCROLL_FAST

def process_arguments():
    """Process the command line arguments and return them as a BoardData object"""
    logger = get_logger()
    board_data = default_board_data

    argument_list = sys.argv[1:]
    options = 'm:g:s:w:'
    long_options = ['message=', 'graphical=', 'scroll=', 'wrap=']
    try:
        # args, vals
        args = getopt.getopt(argument_list, options, long_options)
        if len(args[0]) > 0:
            for arg, val in args[0]:
                if arg in ('-m', '--message'):
                    board_data.message = val
                elif arg in ('-g', '--graphical'):
                    if val == 'True':
                        board_data.graphical = True
                    elif val == 'False':
                        board_data.graphical = False
                    else:
                        logger.warning('Could not parse "graphical" argument: %s', val)
                elif arg in ('-s', 'scroll'):
                    if val in ('slow', 'medium', 'fast'):
                        board_data.scroll_speed = val
                        if val == 'slow':
                            board_data.scroll_wait = SCROLL_FAST
                        elif val == 'medium':
                            board_data.scroll_wait = SCROLL_MED
                        else: # fast
                            board_data.scroll_wait = SCROLL_SLOW
                    else:
                        logger.warning('Invalid scroll speed: %s', val)
                elif arg in ('-w', 'wrap'):
                    if val == 'True':
                        board_data.should_wrap = True
                    elif val == 'False':
                        board_data.should_wrap = False
                    else:
                        logger.warning('Could not parse "wrap" argument: %s', val)
        # --- Verify OS for graphical/hardware
        if on_pi() and board_data.graphical:
            logger.warning('This code cannot be run in graphical' \
                            ' mode on a Raspberry Pi, setting graphical to False')
            board_data.graphical = False
        if not on_pi() and not board_data.graphical:
            logger.warning('This code cannot be run in hardware mode when not run'\
            ' on a Raspberry Pi, setting graphical to True')
            board_data.graphical = True
        # --- Done verifying
        logger.info('message set to: %s', board_data.message)
        logger.info('graphical set to: %s', board_data.graphical)
        logger.info('scroll speed set to: %s (%s)', board_data.scroll_speed, board_data.scroll_wait)
        logger.info('wrap set to: %s', {board_data.should_wrap})

    except getopt.error as err:
        logger.error('getopt error: %s', err)

    return board_data


def main(events=None, initialize_logger=False):
    """Make a very simple display"""

    if initialize_logger:
        init_logger()
    log_os()

    neopolitan = Neopolitan(board_data=process_arguments(), events=events)
    neopolitan.loop()
    del neopolitan

if __name__ == '__main__': # todo: is this still true when running from the thread?
    main(initialize_logger=True) # try False maybe? for testing



"""Board test"""

from neopolitan.board_functions.board import Board

def test_board_scroll():
    """Test that scrolling the data works correctly"""

    initial_data = [1,0,1,0,1,0]
    scroll_once_with_wrap_height_1 = [0,1,0,1,0,1]
    then_scroll_once_without_wrap_height_1 = [1,0,1,0,1]
    board = Board(6)
    board.set_data(initial_data)
    board.scroll(height=1)
    assert board.data == scroll_once_with_wrap_height_1,  \
        'Scrolled data did not match expected'
    board.scroll(wrap=False, height=1, pad=False)
    assert board.data == then_scroll_once_without_wrap_height_1, \
        'Scrolled data did not match expected'
    


"""Draws a board"""

import logging
import pygame
from neopolitan.display.abstract_board_display import BoardDisplay
from neopolitan.const import WIDTH, HEIGHT
# from board import Board

class GraphicalBoardDisplay(BoardDisplay):
    """Draws board data"""
    width = 0
    height = 0
    board = None

    def __init__(self, board, width=WIDTH, height=HEIGHT):
        # pylint: disable=line-too-long
        # pylint: disable=consider-using-f-string
        super().__init__(board)
        assert board.size == width * height, 'board size ({0} does not meet given dimensions {1}x{2}'.format(board.size, width, height)
        self.width = width
        self.height = height

    # pylint: disable=arguments-differ
    def draw_board(self, screen, space, size):
        """Draw all the 'lights' in the board"""
        assert self.board, 'No board assigned'

        for i in range(self.width * self.height):
            if i >= len(self.board.data):
                logging.warning('index %i outside of data array bounds', i)
                return
            color = self.board.data[i]
            row = self.get_row(i)
            col = self.get_col(i)

            if not self.board.data[i]:
                continue
            pos = (int(col * space + space/2), int(row * space + space/2))
            pygame.draw.circle(screen, color, pos, size)

    def get_row(self, idx):
        """Gets the row # from the index"""
        return idx % self.height
    def get_col(self, idx):
        """Gets the col # from the idx"""
        return idx // self.height
    


"""
Handles taking board data and running the
necessary code to get LEDs to show up

This means running 'pygame.draw(...)' or 'pixels.turnOn(...)'
"""

from abc import ABC, abstractmethod

# pylint: disable=too-few-public-methods
class BoardDisplay(ABC):
    """Utilizes board data to turn on/off LEDs"""

    def __init__(self, board):
        # todo: error checking on board
        self.board = board

    @abstractmethod
    def draw_board(self):
        """Use board data to turn on/off individual LEDs"""



"""
Handles initialization of the framework
that will allow LEDs to be displayed

This means initializing the pygame or neopixel libraries
"""

from abc import ABC, abstractmethod

class Display(ABC):
    """Wrapper code for initializing a display framework"""
    def __init__(self):
        self.should_exit = False # todo??

    @abstractmethod
    def draw(self):
        """Initiate the drawing of the board data"""

    @abstractmethod
    def loop(self):
        """Execution loop; used for drawing"""



"""Predefined colors"""

# pylint: disable=fixme
# Todo: use a color library?

WHITE   = (255,255,255)
GRAY    = ( 50, 50, 50)

RED     = (255,  0,  0)
ORANGE  = (255,125,  0)
YELLOW  = (255,255,  0)
GREEN   = (  0,255,  0)
BLUE    = (  0,  0,255)
PURPLE  = (255,  0,255)

ON = WHITE
OFF = GRAY



"""Tests data transformation"""

# pylint: disable=wildcard-import
# pylint: disable=unused-wildcard-import
# pylint: disable=fixme

from neopolitan.writing.letters_8 import *
from neopolitan.writing.data_transformation import frame_length
from neopolitan.writing.data_transformation import symbol_to_array

def test_frame_length():
    """Ensure that frame_length returns the correct length for given arrays"""

    # Todo: more tests necessary?
    assert frame_length(a) == 32, '"a" should take up 32 indices'
    assert frame_length(e) == 32, '"e" should take up 32 indices'
    assert frame_length(i) == 8, '"i" should take up 8 indices'
    assert frame_length(DOLLAR) == 40, '"$" should take up 40 indices'
    assert frame_length('space') == 16, '"space" should take up 16 indices'


def test_symbol_to_array():
    """Test that symbol_to_array correctly calculates array form"""

    # Use '1' instead of color for easier reading
    a_array = [0,0,0,0,0,1,0,0, 0,0,1,0,1,0,1,0, 0,0,1,0,1,0,1,0, 0,0,0,1,1,1,1,0]
    assert symbol_to_array(a, color=1, off=0) == a_array, 'Data array does not match expected'

    # Todo: more tests?
 


"""Handles LED board initialization and cleanup"""
# pylint: disable=import-error
# todo: handle tests in repo
import board as pinout # todo: make sure no import errors
import neopixel
# pylint: disable=no-name-in-module
# todo: why^?
from neopolitan.display.abstract_display import Display
from neopolitan.display.hardware_board_display import HardwareBoardDisplay
from neopolitan.board_functions.board import Board

class HardwareDisplay(Display):
    """Handles LED board initialization and cleanup"""

    def __init__(self, size):

        super().__init__()

        self.size = size

        # Initialize pixels
        self.pixels = neopixel.NeoPixel(pinout.D10, self.size, brightness=0.01, auto_write=False)
        # Initialize board
        self.board_display = HardwareBoardDisplay(Board(size), self.pixels, size)

    def __del__(self):
        """Clean up neopixel"""
        self.pixels.deinit()

    def draw(self):
        """Turn on/off all pixels"""
        self.board_display.draw_board()
        # tell the board to update itself
        self.pixels.show()

    def loop(self):
        """Drawing loop"""
        # todo: handle events

        self.draw()

        # todo: this is never true
        return self.should_exit



import time
from threading import Thread
from queue import Queue

from neopolitan.neop import main
from neopolitan.log import init_logger, get_logger

# todo: doesn't this also need to listen tho??

def add_to_queue(q, e, t):
    time.sleep(t)
    q.put(e)

def runner():
    init_logger()
    get_logger().info('Testing board runner')

    events = Queue()
    t = Thread(target=main, args=(events,))
    t.start()

    Thread(target=add_to_queue, args=(events, 'speed fast', 1,)).start()
    Thread(target=add_to_queue, args=(events, 'speed 1.0', 2,)).start()
    Thread(target=add_to_queue, args=(events, 'speed banana', 3,)).start()
    Thread(target=add_to_queue, args=(events, 'say hello again', 5,)).start()
    # todo: test below
    Thread(target=add_to_queue, args=(events, 'wrap False', 10,)).start()
    Thread(target=add_to_queue, args=(events, 'wrap 1', 15,)).start()
    Thread(target=add_to_queue, args=(events, 'exit', 20,)).start()

    t.join() # no difference

runner()
